import re
from datetime import datetime, timezone
from enum import Enum
from pathlib import Path
from typing import List, Optional, Dict
from uuid import uuid4

import click
import sqlite_utils
from pydantic import BaseModel, root_validator, PrivateAttr, Extra, validator
from pydantic import Field
from rich import filesize


class DetailType(str, Enum):
    ARCHIVE = "archive"
    DEVICE = "device"
    USER = "username"
    LEGAL_HOLD = "legalHold"


class Archive(BaseModel):
    """Class representing an archive."""

    sessionId: Optional[str]
    archiveId: str
    principalId: Optional[str]
    username: Optional[str]
    deviceId: Optional[str]
    deviceName: Optional[str]
    destinationId: Optional[str]
    destinationName: Optional[str]
    storageNodeWebUrl: Optional[str]

    @validator("deviceName")
    def _remove_spaces(cls, value):
        return value.replace(" ", "_")

    @property
    def json_filename(self):
        return (
            f"{self.username}_{self.deviceName}_{self.deviceId}_{self.archiveId}.json"
        )

    @property
    def db_filename(self):
        return f"{self.username}_{self.deviceName}_{self.deviceId}.db"


class FileStatus(str, Enum):
    EXCLUDED = "EXCLUDED"
    SELECTED = "SELECTED"
    IN_PROGRESS = "IN_PROGRESS"
    RETRY = "RETRY"
    COMPLETED = "COMPLETED"


class MacMetadata(BaseModel):
    os: str = "Macintosh"
    creationDate: Optional[int]
    contentModificationDate: Optional[int]
    attributeModificationDate: Optional[int]
    isPackage: Optional[bool]
    isHidden: Optional[bool]
    hasHiddenExtension: Optional[bool]
    labelNumber: Optional[int]
    uid: Optional[int]
    gid: Optional[int]
    mode: Optional[int]
    aclText: Optional[str]
    bsdFlags: Optional[bool]
    specialDeviceId: Optional[int]
    xAttrs: Optional[List[str]]


class LinuxMetadata(BaseModel):
    os: str = "Linux"
    uid: Optional[int]
    gid: Optional[int]
    mode: Optional[int]
    cTime: Optional[int]


class WindowsMetadata(BaseModel):
    os: str = "Windows"
    fileAttributeFlags: Optional[int]
    aclText: Optional[List[str]]
    accessDate: Optional[int]
    createDate: Optional[int]
    changeDate: Optional[int]
    lastModifiedDate: Optional[int]


metadata_map = {
    "Windows": WindowsMetadata,
    "Linux": LinuxMetadata,
    "Macintosh": MacMetadata,
}


class FileMetadata(BaseModel, extra=Extra.ignore):
    archiveId: str
    fileId: str
    backupSetId: Optional[int]
    fileType: str
    path: str
    versionTimeMillis: int
    sourceLastModifiedMillis: int
    sourceLengthBytes: int
    sourceMD5Checksum: Optional[str]
    sourceSHA256Checksum: Optional[str]
    deleted: bool
    deletionTimeMillis: Optional[int]
    status: str = FileStatus.EXCLUDED
    excludeReason: Optional[str]

    @root_validator(pre=True)
    def handle_deleted_file(cls, values: dict):  # noqa
        if values["deleted"]:
            last_non_deleted = values.pop("lastNonDeletedEntry")
            values["deletionTimeMillis"] = values["versionTimeMillis"]
            values.update(last_non_deleted)
        return values

    @validator("status")
    def validate_status(cls, value):
        FileStatus(value)
        return value


class RestorableFile(BaseModel):
    deviceId: str
    archiveId: str
    storageNodeWebUrl: str
    destinationName: str
    fileId: str
    path: str
    versionTimeMillis: int
    sourceLastModifiedMillis: int
    sourceLengthBytes: int
    sourceMD5Checksum: Optional[str]
    sourceSHA256Checksum: Optional[str]
    backupSetId: Optional[int]
    error: Optional[str]
    failure_count: int = 0
    status: FileStatus

    @property
    def restore_path(self):
        # Convert Windows drive to letter folder
        value = re.sub(r"^([A-Z]):", r"\1", self.path)
        # remove root slash "/" if present
        value = value.lstrip("/")
        return f"{self.deviceId}/{value}"

    @property
    def size(self):
        """Human-readable file size. 1KB == 1000 Bytes."""
        return filesize.decimal(self.sourceLengthBytes)


class FileSelection(BaseModel):
    """
    Class that defines how to restore files from an archive.

    **Fields**

    * **includeDeleted**: `bool` Include deleted files in the requested archive metadata. Default=False
    * **backupSetIds**: `Optional[List[int]]` Optional list of backup set Ids to select for restore (V3 archives only).
    * **deletedMaxAge**: `Optional[datetime]` Filter deleted files by age of deletion. Files that were marked
        deleted in the archive prior to provided date will be excluded from restore.
    * **restoreAsOf**: `Optional[datetime]` Restore file versions as of the provided date.
    * **includePatterns**: `List[str]` Optional list of regex patterns to include, any files _not matching_ one of these
        patterns will be excluded from restore.
    * **excludePatterns**: `List[str]` Optional list of regex patterns to exclude, any files _matching_ one of these
        patterns will be excluded from restore.
    """

    includeDeleted: bool = False
    backupSetIds: Optional[List[int]]
    deletedMaxAge: Optional[datetime]
    restoreAsOf: Optional[datetime]
    includePatterns: List[str] = Field(default_factory=list)
    excludePatterns: List[str] = Field(default_factory=list)

    @validator("includePatterns")
    def validate_include_regexes(cls, value):
        try:
            [re.compile(v) for v in value]
            return value
        except re.error as regex_error:
            raise click.ClickException(
                f"Unable to parse regex for include pattern '{regex_error.pattern}': {regex_error}"
            )

    @validator("excludePatterns")
    def validate_exclude_regexes(cls, value):
        try:
            [re.compile(v) for v in value]
            return value
        except re.error as regex_error:
            raise click.ClickException(
                f"Unable to parse regex for exclude pattern '{regex_error.pattern}': {regex_error}"
            )

    @property
    def endTimeMillis(self):
        if self.restoreAsOf is not None:
            return int(self.restoreAsOf.astimezone(timezone.utc).timestamp() * 1000)

    @property
    def deletedAgeMillis(self):
        if self.deletedMaxAge is not None:
            return int(self.deletedMaxAge.astimezone(timezone.utc).timestamp() * 1000)


class Device:
    def __init__(self, username: str, name: str, deviceId: str, db_file: Path):
        self.username = username
        self.name = name
        self.deviceId = deviceId
        self.db = sqlite_utils.Database(db_file)
        self.progress = None
        self.archive_progress = None
        self.task_id = None

    @property
    def completedFiles(self):
        result = next(
            self.db.query(
                """
                SELECT coalesce(count(distinct(fileId)), 0) as completedFiles
                FROM file
                WHERE fileType != 'DIRECTORY'
                AND status = 'COMPLETED'
                """
            )
        )
        return result["completedFiles"]

    @property
    def completedBytes(self):
        result = next(
            self.db.query(
                """
                with distinct_files as (
                    SELECT max(fileId), sourceLengthBytes
                    FROM latest_versions
                    WHERE fileType != 'DIRECTORY'
                    AND status = 'COMPLETED'
                    GROUP BY fileId
                )
                SELECT coalesce(sum(sourceLengthBytes), 0) as completedBytes
                FROM distinct_files
                """
            )
        )
        return result["completedBytes"]

    @property
    def totalFiles(self):
        result = next(
            self.db.query(
                """
                SELECT coalesce(count(distinct(fileId)), 0) as totalFiles
                FROM file
                WHERE fileType != 'DIRECTORY'
                AND status != 'EXCLUDED'
                """
            )
        )
        return result["totalFiles"]

    @property
    def totalBytes(self):
        result = next(
            self.db.query(
                """
                WITH dedupe as (
                    SELECT
                        fileId,
                        sourceLengthBytes,
                        row_number() OVER (PARTITION BY fileId) as num
                    FROM latest_versions
                    WHERE status != 'EXCLUDED'
                )
                SELECT coalesce(sum(sourceLengthBytes), 0) as totalBytes
                FROM dedupe
                WHERE num = 1
                """
            )
        )
        return result["totalBytes"]


class JobDefinition(BaseModel):
    """
    Class that defines a RestoreJob parameters.

    **Fields**

    * **jobId**: `str` Identifier for job (will show in Audit Log entries). If not provided a new UUID will be generated.
    * **type**: `Optional[str]` Optional indicator of how the archives in the job were selected, i.e. by archive,
        device, username, or legal hold ID.
    * **identifier**: `Optional[str]` Optional indicator of what identifiers were provided of the given `type`.
    * **selection**: `FileSelection` Selection criteria for what files to restore from the archive.
    * **targetFolder**: `Path` Directory where the job metadata and restored files should be saved to (defaults to
        current working directory).
    * **zipResults**: `bool` Indicates if the restored files should be compressed to a zip archive per device in the job.
    """

    jobId: str = Field(default_factory=lambda: str(uuid4()))
    type: Optional[str]
    identifier: Optional[str]
    selection: FileSelection = Field(default_factory=FileSelection)
    targetFolder: Path = Field(default_factory=lambda: Path().absolute())
    zipResults: bool = False
    archives: List[Archive]
    totalFileCount: int = 0
    totalByteCount: int = 0
    _devices: Dict[str, Device] = PrivateAttr(default_factory=dict)

    @property
    def filename(self):
        return f"restore_{self.jobId}.json"

    @property
    def folder(self):
        return self.targetFolder / self.jobId



from datetime import datetime
from datetime import timedelta
from typing import Generator, Optional

import httpx
from pydantic import BaseModel
from pydantic import PrivateAttr


class AuthToken(BaseModel):
    token_type: str
    expires_in: int
    access_token: str
    _init_time: datetime = PrivateAttr(default_factory=datetime.utcnow)

    @property
    def expired(self):
        elapsed = datetime.utcnow() - self._init_time
        return elapsed > timedelta(seconds=self.expires_in)


class APIClientAuth(httpx.Auth):
    def __init__(self, url: str, api_client_id: str, api_client_secret: str):
        self.url = url
        self.api_client_id = api_client_id
        self.api_client_secret = api_client_secret
        self.token: Optional[AuthToken] = None

    def auth_flow(self, request) -> Generator[httpx.Request, httpx.Response, None]:
        request.headers["Authorization"] = f"Bearer {self.get_token()}"
        response = yield request
        if response.status_code == 401:
            request.headers["Authorization"] = f"Bearer {self.get_token()}"
            yield request

    def get_token(self) -> str:
        if self.token is None or self.token.expired:
            response = httpx.post(
                url=f"{self.url}/api/v3/oauth/token?grant_type=client_credentials",
                auth=(self.api_client_id, self.api_client_secret),
            )
            response.raise_for_status()
            self.token = AuthToken(**response.json())
        return self.token.access_token



import re
from pathlib import Path
from queue import Queue
from typing import Dict
from typing import Generator
from typing import List

import sqlite_utils

from bulk_restore_tool.models import Archive
from bulk_restore_tool.models import FileMetadata
from bulk_restore_tool.models import FileSelection
from bulk_restore_tool.models import RestorableFile
from bulk_restore_tool.models import FileStatus


LATEST_VERSION_VIEW_SQL = """
    WITH latest_versions AS (
        SELECT fileId, max(sourceLastModifiedMillis) as sourceLastModifiedMillis
        FROM file
        GROUP BY fileId
    )
    SELECT
        f.archiveId,
        f.fileId,
        f.backupSetId,
        f.fileType,
        f.path,
        f.versionTimeMillis,
        f.sourceLastModifiedMillis,
        f.sourceLengthBytes,
        f.sourceSHA256Checksum,
        f.sourceMD5Checksum,
        f.deleted,
        f.deletionTimeMillis,
        f.status,
        f.excludeReason
    FROM file f
    JOIN latest_versions v on (f.fileId = v.fileId)
    WHERE f.sourceLastModifiedMillis = v.sourceLastModifiedMillis
"""


RESTORABLE_VIEW_SQL = """
    SELECT
        a.deviceId,
        a.archiveId,
        a.destinationName,
        a.storageNodeWebUrl,
        f.fileId,
        f.path,
        f.backupSetId,
        f.versionTimemillis,
        f.sourceLastModifiedMillis,
        f.sourceLengthBytes,
        f.sourceSHA256Checksum,
        f.sourceMD5Checksum,
        f.status
    FROM latest_versions f
    JOIN archive a on (a.archiveId = f.archiveId)
    WHERE f.fileType != 'DIRECTORY'
    AND f.status = 'SELECTED'
"""


def select_files(
    archiveId: str, files: List[Dict], selection: FileSelection
) -> Generator[Dict, None, None]:
    """
    Marks files as 'selected' if they meet file selection criteria.
    If not selected, the 'excludeReason' field will be populated.
    """
    include_patterns = [re.compile(pat) for pat in selection.includePatterns]
    exclude_patterns = [re.compile(pat) for pat in selection.excludePatterns]
    for file in files:
        model = FileMetadata(**file, archiveId=archiveId)

        # apply include patterns
        if include_patterns:
            for regex in include_patterns:
                if regex.match(model.path):
                    model.status = FileStatus.SELECTED
                    model.excludeReason = None
        else:
            model.status = FileStatus.SELECTED
            model.excludeReason = None

        # skip other filter logic if file isn't included
        if not model.status == FileStatus.SELECTED:
            model.excludeReason = "Didn't match any include patterns."
            yield model.dict()

        # apply exclude patterns
        if exclude_patterns:
            for regex in exclude_patterns:
                if regex.match(model.path):
                    model.status = FileStatus.EXCLUDED
                    model.excludeReason = f"Matched exclude pattern: {regex.pattern}"

        # apply backup set filter
        if selection.backupSetIds:
            if model.backupSetId is None:
                raise V1BackupSetException(
                    f"{model.archiveId} is a V1 archive and backup sets are not accessible for restores. Unable to filter by backup set ID."
                )
            if model.backupSetId not in selection.backupSetIds:
                model.status = FileStatus.EXCLUDED
                model.excludeReason = (
                    f"Not in selected backupSetIds: {selection.backupSetIds}"
                )

        # apply deleted file age filter
        if selection.includeDeleted and selection.deletedAgeMillis is not None:
            if model.deleted and model.deletionTimeMillis < selection.deletedAgeMillis:
                model.status = FileStatus.EXCLUDED
                model.excludeReason = (
                    f"Delete prior to maxDeletedAge: {selection.deletedMaxAge}"
                )
        yield model.dict()


def save_archive_metadata_sqlite(
    archive: Archive,
    files: List[dict],
    target: Path,
    selection: FileSelection,
) -> Path:
    file_path = target / archive.db_filename
    db = sqlite_utils.Database(file_path)
    archive_table = db["archive"].create(  # type: ignore
        columns={name: field.type_ for name, field in Archive.__fields__.items()},
        pk="archiveId",
        if_not_exists=True,
    )
    file_table = db["file"].create(  # type: ignore
        columns={name: field.type_ for name, field in FileMetadata.__fields__.items()},
        pk=("archiveId", "fileId"),
        if_not_exists=True,
    )
    archive_table.upsert(archive.dict(), pk="archiveId")
    selected_files = select_files(
        archiveId=archive.archiveId, files=files, selection=selection
    )
    file_table.upsert_all(selected_files, pk=("archiveId", "fileId"))

    db.create_view("latest_versions", LATEST_VERSION_VIEW_SQL, replace=True)
    db.create_view("restorable", RESTORABLE_VIEW_SQL, replace=True)
    return file_path


def restore_queue(db: sqlite_utils.Database, archive_id: str) -> Queue:
    queue: Queue = Queue()
    for file in db.query(
        """
        SELECT *
        FROM restorable
        WHERE archiveId = :archiveId
        ORDER BY sourceLengthBytes ASC
        """,
        {"archiveId": archive_id},
    ):
        queue.put(RestorableFile(**file))
    return queue


def get_file_status(file: RestorableFile, db: sqlite_utils.Database) -> FileStatus:
    result = db.execute(
        sql="select status from file where fileId = :fileId and archiveId = :archiveId",
        parameters={"fileId": file.fileId, "archiveId": file.archiveId},
    ).fetchone()
    return FileStatus(result[0])


def set_file_status(
    file: RestorableFile, status: FileStatus, db: sqlite_utils.Database
):
    with db.conn:
        db.execute(
            sql="update file set status = :status where fileId = :fileId",
            parameters={"status": status, "fileId": file.fileId},
        )


class V1BackupSetException(Exception):
    ...



from pathlib import Path

import click
from rich.console import Console

from bulk_restore_tool.__about__ import __version__
from bulk_restore_tool._client import Client
from bulk_restore_tool._db import V1BackupSetException
from bulk_restore_tool._job import RestoreJob
from bulk_restore_tool.models import DetailType
from bulk_restore_tool.models import JobDefinition
from bulk_restore_tool.models import FileSelection

console = Console(stderr=True)


def store_in_context(ctx, param, arg):
    """
    Callback for `click.option`s that stores option value in the `ctx.obj` (also ensures this is a dict).

    Useful for enabling "global" options that can be passed at any stage of the command or sub-commands.
    """
    if arg:
        ctx.ensure_object(dict)
        ctx.obj.update({param.name: arg})
    return arg


log_file_option = click.option(
    "--log-file",
    help="Specify location where logging gets written. Can also be set by BRT_LOG_FILE environment variable. "
    "Default='bulk_restore.log' in current directory.",
    callback=store_in_context,
    expose_value=False,
)
log_level_option = click.option(
    "--log-level",
    type=click.Choice(["CRITICAL", "WARN", "INFO", "DEBUG"]),
    help="Specify logging level for main logger. Can also be set by BRT_LOG_LEVEL environment variable. Defaults to INFO.",
    callback=store_in_context,
    expose_value=False,
)


def logging_options(f):
    f = log_file_option(f)
    f = log_level_option(f)
    return f


@click.group(
    context_settings={
        "help_option_names": ["-h", "--help"],
        "max_content_width": 200,
    },
    invoke_without_command=True,
    no_args_is_help=True,
)
@click.option(
    "--version", is_flag=True, help="Print the currently installed version of the tool."
)
@logging_options
def brt(version, **kwargs):
    """
    \b
    Code42 Bulk Restore Tool

    To authenticate, set the following variables either in your shell environment or an .env file in your
    current working directory:

    \b
        BRT_URL
        BRT_API_CLIENT_ID
        BRT_API_CLIENT_SECRET

    For details on API Client authentication, see: https://support.code42.com/Incydr/Admin/Code42_console_reference/API_clients
    """
    if version:
        print(__version__)


@brt.command()
@click.argument("id_", metavar="ID")
@click.option(
    "--name",
    help="Specify a name for the restore job. This name will be visible in audit log entries for the restore session. "
    "Default = A generated UUID.",
)
@click.option(
    "--type",
    "type_",
    type=click.Choice([d.value for d in DetailType]),
    default="archive",
    help="Specify what type of identifier to gather archive metadata for. Default = archive",
)
@click.option(
    "--target-folder",
    type=click.Path(),
    help="Specify target folder where archive metadata will be stored (will be created if it doesn't exist). "
    "Default = Directory `brt` is executed in.",
)
@click.option(
    "--definition-folder",
    type=click.Path(),
    help="Specify target folder the job definition .json file will be written to. "
    "Default = Directory `brt` is executed in.",
)
@click.option(
    "--zip",
    "zip_",
    is_flag=True,
    help="Restore files into .zip archives, one per device restored.",
)
@click.option(
    "--include-pattern",
    multiple=True,
    type=str,
    help="Specify regex pattern to select which file paths to include in the restore. Can be passed multiple times.",
)
@click.option(
    "--exclude-pattern",
    multiple=True,
    type=str,
    help="Specify regex pattern to select which file paths to exclude in the restore. Can be passed multiple times. A "
    "match on an exclude pattern will take precedence over any matching include patterns.",
)
@click.option(
    "--include-deleted",
    is_flag=True,
    help="Include files that have been marked as deleted in the archive file metadata. Default = false",
)
@click.option(
    "--max-deleted-date",
    type=click.DateTime(),
    help="Specify date for maximum age of deleted files to restore. Any files that were marked as deleted before this "
    "date will be excluded from restore. Default = None",
)
@click.option(
    "--as-of-date",
    type=click.DateTime(),
    help="Restore files as of specified date. Default = None",
)
@click.option(
    "--backup-set-ids",
    type=int,
    multiple=True,
    help="(V3 archives only) Specify which backup sets from the archive should be included, can be passed multiple "
    "times to include files from multiple sets. Default = Include files from all backup sets.",
)
@click.option(
    "--no-calculate",
    is_flag=True,
    help="Skip gathering of archive metadata and calculation of restore totals and only create job file. "
    "Default = false",
)
@logging_options
@click.pass_context
def create_job(
    ctx,
    id_,
    target_folder,
    definition_folder,
    type_,
    zip_,
    include_pattern,
    exclude_pattern,
    include_deleted,
    backup_set_ids,
    max_deleted_date,
    as_of_date,
    name,
    no_calculate,
):
    """
    Creates a restore job for a set of archives.

    Archives can be included in the job by specifying the identifier type (archive, device, username, or legalHold)
    via the --type option (defaults to archive) and providing a comma-separated list of identifiers as the main
    command-line argument. Some examples:

        \b
        brt create-job --type archive id1,id2,id3
        brt create-job --type username user@example.com
        brt create-job --type device 1234,4567
        brt create-job --type legalHold 7890

    If the --name option is not provided, a new UUID will be created as the job name and the parameters for the job will
    be saved in restore_<job_name>.json

    A <job_name> folder will also be created in the --target-folder path to store the archive metadata and restored
    files (if --target-folder option is not provided, it defaults to the current working directory).

    In the <job_name> folder, a human-readable .json file will be generated for each archive containing the file
    metadata from that archive. Additionally, a sqlite .db file will also be generated for each device in the job to
    track restore progress per device. Because a device can have multiple (destination) archives and/or multiple backup
    sets within a single archive, it's possible that some file paths will be duplicated within the full set of a single
    device's metadata. While duplicates will remain in the individual archive metadata files, they will be de-duplicated
    when restoring and when calculating the total file count and byte count of the restore job (the fileId with the most
    recent modification timestamp will be restored).
    """
    target_folder = Path(target_folder) if target_folder else Path(".")
    params = {
        "includeDeleted": include_deleted,
        "backupSetIds": backup_set_ids,
        "deletedMaxAge": max_deleted_date,
        "restoreAsOf": as_of_date,
        "includePatterns": include_pattern,
        "excludePatterns": exclude_pattern,
    }
    log_options = ctx.obj or {}
    client = Client(**log_options)
    ids = id_.split(",")
    archives = []
    for _id in ids:
        archives.extend(client.get_archive_details(_id, id_type=type_))
    job_definition = JobDefinition(
        type=type_,
        identifier=_id,
        selection=FileSelection(**params),
        archives=archives,
        targetFolder=target_folder.absolute(),
        zipResults=zip_,
    )
    if name:
        job_definition.jobId = name
    definition_folder = Path(definition_folder) if definition_folder else Path(".")
    definition_file = definition_folder / job_definition.filename
    with open(definition_file, "w") as f:
        f.write(job_definition.json(indent=2))
    if no_calculate:
        return
    job = RestoreJob(definition=job_definition)
    try:
        job.prepare()
        console.print(f"Job definition saved to: {definition_file}")
    except V1BackupSetException as err:
        console.print(f"[red]Error:[/red] {err.args[0]}")


@brt.command()
@click.option(
    "--quiet",
    "-q",
    is_flag=True,
    help="Silence output to stdout. Restore progress can be tracked via log files.",
)
@click.argument("JOB_FILE", type=click.File())
@logging_options
@click.option(
    "--device-log-level",
    type=click.Choice(["CRITICAL", "ERROR", "WARN", "INFO", "DEBUG"]),
    help="Specify level for loggers that write to the individual <deviceGuid>.log and <deviceGuid>_error.log files in "
    "the restore target directory. Can also be set by BRT_DEVICE_LOG_LEVEL environment variable. Defaults to INFO. "
    "Setting to CRITICAL effectively disables all device-specific logging.",
    callback=store_in_context,
    expose_value=False,
)
@click.pass_context
def start_job(ctx, job_file, quiet):
    """Begins restoring from archives defined in a given job file."""
    log_options = ctx.obj or {}
    job_definition = JobDefinition.parse_raw(job_file.read())
    job = RestoreJob(definition=job_definition, quiet=quiet, **log_options)
    try:
        job.start()
    except V1BackupSetException as err:
        console.print(f"[red]Error:[/red] {err.args[0]}")
